# encoding: utf-8
import argparse
import os
import re
from PIL import Image
import numpy as np
import logging
logging.basicConfig(format='%(asctime)s-%(levelname)s-%(name)s-   %(message)s',
                    level=logging.INFO)

logger = logging.getLogger(__name__)

###################
# 找边框区域，仅仅使用diff_1的设置参数， 使用灰度图片
###################
border_max_len_rate = 0.3  # 边框长度不能超过该值，否则不认为是边框
diff_threshold = 15  # 差值图像转为二值图像，阈值
diff_count_threshold = 0.015  # 如果像素线中（梯度！=0）的数量不超过该值，则认为是边框的一部分，否则不属于边框
more_pix_count = 5  # 可能裁剪后还存留稍许的边界线，这里多裁剪几个像素
remain_area_threshold = 0.5  # 如果裁剪面积超过50%，则不做裁剪

# (modify_is_border_1)如果当前像素点不是边框点，但它前后窗口大小内都是边框点，则该点也是边框点
diff_count_upper_threshold = 0.8  # 如果像素差值不为0的比例超过该值时，则直接认为该线不能作为边框区域的一部分；否则接着判断下面的阈值
window_rate_threshold = 1  # 窗口是边框的像素比例阈值(modify_is_border_1)
false_pixes_modify_window_size = 3  # 如果当前像素点不是边框点，但它前后窗口大小内都是边框点，则该点也是边框点(modify_is_border_1)

continuous_is_border_min_count = 6  # 连续出现的是边框像素点的数量阈值(modify_is_border_2)

# 存在找到的边框区域，可能在垂直方向上不是纯色，即方差（或标准差）较大，则不认为是边框区域
border_region_std_limit = False
border_region_std_threshold = 0.95  # 标准差小于该值，才可认为是边框区域

# 是否使用突变边框线(边框区域和图内容的分界线)兜底
mutation_limit = True  # 是否使用突变边框线兜底
mutation_limit_window_size = 0.01  # 在窗口大小范围内找突变边框线
huge_diff_threshold = 0.2  # 当（梯度！=0）像素点的数量超过该阈值时，认为时突变边框线；否则不是

###################
# 找边框区域和图内容的分界线
###################
boundary_limit = False
boundary_diff_threshold = 10  # 差值图像转为二值图像，阈值
boundary_diff_count_threshold = 0.75  # 如果像素线中（梯度！=0）的数量超过该值，则认为可能是分界线，否则不是


def modify_is_border_1(is_border, diff_count, ref_len):
    """
    在窗口范围内，如果是边框的像素比例超过阈值时，则认为该像素点是边框的一部分，否则不是；
    只针对不是边框像素点做处理，如果是边框像素点则不管
    这个函数是针对很小一部分不是边框像素点，做处理的
    例如: comic/3415/233883415.jpg
    """
    # window_rate_threshold = 0.8  # 窗口是边框的像素比例阈值
    window_size = 3
    # 假设： 开头是边框，结尾不是边框
    pad_is_border = np.concatenate(
        (
            np.array([True] * window_size, dtype=bool),
            is_border,
            np.array([False] * window_size, dtype=bool),
        ),
        axis=0
    )
    final_is_border = []
    for i in range(window_size, pad_is_border.shape[0] - window_size):
        if pad_is_border[i]:
            final_is_border.append(True)
        elif diff_count[i-window_size] / ref_len > diff_count_upper_threshold:  # 像素差不为0的数量超过该比例了，则不能认为是边框区域的一部分
            final_is_border.append(False)
        elif pad_is_border[i-window_size:i+window_size+1].sum() / (window_size * 2) >= window_rate_threshold:
            final_is_border.append(True)
        else:
            final_is_border.append(False)

    final_is_border = np.array(final_is_border)

    assert final_is_border.shape == is_border.shape

    return final_is_border


def get_all_border_index(is_border, diff):
    border_indexes = [0]
    continuous_is_border_count = 0
    for i in range(is_border.shape[0]):
        if is_border[i]:
            continuous_is_border_count += 1
        else:
            if continuous_is_border_count >= continuous_is_border_min_count:
                if border_region_std_limit:  # 是否边框区域方差检查
                    std = np.std(
                        np.mean(
                            diff[i - continuous_is_border_count:i, :],
                            axis=0
                        )
                    )
                    # print(std)
                    if std < border_region_std_threshold:
                        border_indexes.append(i)
                else:
                    border_indexes.append(i)

            continuous_is_border_count = 0  # 进行下一次的重新计数

    return border_indexes


def get_border_len(diff, ref_len_1, ref_len_2):
    """
    索引值大的边框索引，且附近存在明显的边框线
    """
    H, W = diff.shape
    binary_diff = (diff > diff_threshold).astype(np.int8)
    diff_count = binary_diff.sum(axis=1)
    is_border = diff_count <= ref_len_1 * diff_count_threshold

    # import pdb;pdb.set_trace()
    is_border = modify_is_border_1(is_border, diff_count, W)

    border_indexes = get_all_border_index(is_border, diff)

    if len(border_indexes) == 1:
        return border_indexes[0]  # 必然是0
    else:
        if mutation_limit:  # 是否使用突变边框线(边框区域和图内容的分界线)兜底
            for bi in border_indexes[::-1]:
                exist_border_line = False
                for i in range(
                        max(0, bi - int(ref_len_2 * mutation_limit_window_size)),
                        min(bi + int(ref_len_2 * mutation_limit_window_size), H)
                ):
                    if diff_count[i] / W > huge_diff_threshold:
                        exist_border_line = True
                        break
                if exist_border_line:
                    return bi
            return border_indexes[0]  # 必然是0
        else:
            return border_indexes[-1]


def get_diff_1(img_np):
    """
    差分，从外向内的方向
    """
    H, W = img_np.shape
    cur_use_len_h = min(int(H * border_max_len_rate) + 2, H - 1)
    cur_use_len_w = min(int(W * border_max_len_rate) + 2, W - 1)

    # top
    top_diff = np.absolute(
        img_np[:cur_use_len_h, :] - img_np[1:cur_use_len_h + 1, :]
    )

    # bottom
    bottom_diff = np.absolute(
        np.flip(
            img_np[-cur_use_len_h:, :] - img_np[-(cur_use_len_h + 1):-1, :],
            axis=0
        )
    )

    # left
    left_diff = np.absolute(
        (img_np[:, :cur_use_len_w] - img_np[:, 1:cur_use_len_w + 1]).T
    )

    # right
    right_diff = np.absolute(
        np.flip(
            img_np[:, -cur_use_len_w:] - img_np[:, -(cur_use_len_w + 1):-1],
            axis=1
        ).T
    )

    diff = {
        'top': top_diff,
        'bottom': bottom_diff,
        'left': left_diff,
        'right': right_diff
    }
    return diff


def get_diff_2(img_np):
    """
    差分，垂直于(从外向内的方向)
    """
    H, W = img_np.shape
    cur_use_len_h = min(int(H * border_max_len_rate) + 2, H - 1)
    cur_use_len_w = min(int(W * border_max_len_rate) + 2, W - 1)

    # top
    top_diff = np.absolute(
        img_np[:cur_use_len_h, :-1] - img_np[:cur_use_len_h, 1:]
    )
    top_diff = np.concatenate(
        (top_diff, top_diff[:, -1:]),
        axis=1
    )

    # bottom
    bottom_diff = np.absolute(
        np.flip(
            img_np[-cur_use_len_h:, :-1] - img_np[-cur_use_len_h:, 1:],
            axis=0
        )
    )
    bottom_diff = np.concatenate(
        (bottom_diff, bottom_diff[:, -1:]),
        axis=1
    )

    # left
    left_diff = np.absolute(
        (img_np[:-1, :cur_use_len_w] - img_np[1:, :cur_use_len_w]).T
    )
    left_diff = np.concatenate(
        (left_diff, left_diff[:, -1:]),
        axis=1
    )

    # right
    right_diff = np.absolute(
        np.flip(
            img_np[:-1, -cur_use_len_w:] - img_np[1:, -cur_use_len_w:],
            axis=1
        ).T
    )
    right_diff = np.concatenate(
        (right_diff, right_diff[:, -1:]),
        axis=1
    )

    diff = {
        'top': top_diff,
        'bottom': bottom_diff,
        'left': left_diff,
        'right': right_diff
    }
    return diff


def get_final_diff(img_np, only=True):
    diff_1 = get_diff_1(img_np)
    diff_2 = get_diff_2(img_np)

    diff = {
        'top': diff_1['top'] if only else diff_1['top'] + diff_2['top'],
        'bottom': diff_1['bottom'] if only else diff_1['bottom'] + diff_2['bottom'],
        'left': diff_1['left'] if only else diff_1['left'] + diff_2['left'],
        'right': diff_1['right'] if only else diff_1['right'] + diff_2['right']
    }

    return diff


def get_boundary_len(diff, limit_len_1, limit_len_2):
    """
    (1) 像素线中（梯度！=0）的数量超过(limit_len_1 * diff_count_threshold)，
        才认为可能是分界线，否则不是
    (2) 边框长度不能超过(limit_len_2 * border_max_len)
    """
    binary_diff = (diff > boundary_diff_threshold).astype(np.int8)
    diff_count = binary_diff.sum(axis=1)
    is_border_line = diff_count > limit_len_1 * boundary_diff_count_threshold

    border_len = 0
    for i in range(0, min(is_border_line.shape[0], int(limit_len_2 * border_max_len_rate))):
        if is_border_line[i]:
            border_len = i
    return border_len


def get_real_image_index(img_np):
    H, W = img_np.shape
    diff = get_final_diff(img_np)

    # import pdb;pdb.set_trace()
    # top
    top_border_len = get_border_len(diff['top'], W, H)
    if boundary_limit:
        top_boundary_len = get_boundary_len(diff['top'], W, H)
        if top_border_len == 0:
            top_border_len = top_boundary_len
    # bottom
    bottom_border_len = get_border_len(diff['bottom'], W, H)
    if boundary_limit:
        bottom_boundary_len = get_boundary_len(diff['bottom'], W, H)
        if bottom_border_len == 0:
            bottom_border_len = bottom_boundary_len
    # left
    left_border_len = get_border_len(diff['left'], H, W)
    if boundary_limit:
        left_boundary_len = get_boundary_len(diff['left'], H, W)
        if left_border_len == 0:
            left_border_len = left_boundary_len
    # right
    right_border_len = get_border_len(diff['right'], H, W)
    if boundary_limit:
        right_boundary_len = get_boundary_len(diff['right'], H, W)
        if right_border_len == 0:
            right_border_len = right_boundary_len

    # 可能裁剪后还存留稍许的边界线，这里多裁剪几个像素
    top_border_len = top_border_len + more_pix_count if top_border_len > 0 else top_border_len
    bottom_border_len = bottom_border_len + more_pix_count if bottom_border_len > 0 else bottom_border_len
    left_border_len = left_border_len + more_pix_count if left_border_len > 0 else left_border_len
    right_border_len = right_border_len + more_pix_count if right_border_len > 0 else right_border_len

    index = {
        'height': (top_border_len, img_np.shape[0] - bottom_border_len),
        'width': (left_border_len, img_np.shape[1] - right_border_len)
    }

    # 如果裁剪后面积小于remain_area_threshold，则不做裁剪
    remain_area_rate = \
        (
            (index['height'][1] - index['height'][0] + 1) *
            (index['width'][1] - index['width'][0] + 1)
        ) / (H * W)
    if remain_area_rate < remain_area_threshold:
       index = {'height': (0, H), 'width': (0, W)}

    # 存在把所有的内容都裁剪掉，这里修正
    # 例如: comic/1522/295761522.jpg
    if index['height'][0] >= index['height'][1]:
        index['height'] = (0, img_np.shape[0])
    if index['width'][0] >= index['width'][1]:
        index['width'] = (0, img_np.shape[1])

    return index


def remove_image_border(in_dir, out_dir):
    logger.info('begin')
    grey_flag = True  # 使用灰度图片
    images_count = 0
    for root, dirs, files in os.walk(in_dir):
        for file in files:
            #if file.endswith('.dup'):
            #    continue
            try:
                im = Image.open(os.path.join(root, file))
                orig_im_np = np.asanyarray(im)
            except Exception as e:
                print('%s\t%s' % (os.path.join(root, file), e))
                continue

            if grey_flag:
                im_grey = im.convert('L')
                # im_grey.save(os.path.join(args.out_dir, 'grey_' + file))

                im_np = np.asanyarray(im_grey)
                # 必须做类型转换，否则对于原始类型uint8时，0-1=255的
                im_np = im_np.astype(np.int16)
            else:
                if len(orig_im_np.shape) == 3:
                    im_np = np.mean(orig_im_np, axis=2)
                else:
                    im_np = orig_im_np

            img_index = get_real_image_index(im_np)
            # print('%s: %s, %s' % (os.path.join(root, file), orig_im_np.shape, img_index))
            # import pdb;pdb.set_trace()
            real_im_np = orig_im_np[
                img_index['height'][0]:img_index['height'][1],
                img_index['width'][0]:img_index['width'][1]
            ]
            real_im = Image.fromarray(real_im_np)

            final_out_dir = os.path.join(out_dir, re.sub('^/*', '', root.replace(in_dir, '')))
            if not os.path.exists(final_out_dir):
                os.makedirs(final_out_dir, exist_ok=True)
            real_im.save(os.path.join(final_out_dir, file))

            images_count += 1
            if images_count % 1000 == 0:
                logger.info(images_count)

    logger.info('%d end' % images_count)


def parse_args():
    parser = argparse.ArgumentParser(description="")
    parser.add_argument("--in_dir", help="")
    parser.add_argument("--out_dir", help="")
    args = parser.parse_args()
    return args


if __name__ == "__main__":
    args = parse_args()
    remove_image_border(args.in_dir, args.out_dir)


